// bot4.c
// The bot to post frames to the OTT
// 08.04.2015
// 
// Copyright (C) 2015  Balthasar Szczepański
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/wait.h>

#define OK              0
#define POST_MSTD     256
#define INDELIVERABLE 512
#define NOPOSTS         3
#define STOP_MSTD       4
#define NAME_MSTD       5
#define DOWN_MSTD       6
#define PREP_MSTD       7
#define LAST_MSTD       8
#define ONGD_MSTD       9

#define LINE_LIMIT 16000

#define USERAGENT "bothasar_p (http://1190.bicyclesonthemoon.info/ongoing.htm; BFTF post bot)"

#define OTT_ADDRESS "http://forums.xkcd.com/viewtopic.php?f=7&t=101043&start="
#define TTO_ADDRESS "http://1190.bicyclesonthemoon.info/ott/view?start="
#define MSTD_ADDRESS "http://1190.bicyclesonthemoon.info/ott/mview?start="
#define LASTPOST_ADDRESS "http://forums.xkcd.com/viewtopic.php?f=7&t=101043&sd=d"

#define LAST_PATH "/eizm/mem/ong1/lasttime"
#define LOG_PATH "/eizm/log/ong1/bot4.log"
#define STOP_PATH "/eizm/tmp/ong1/bot4stop"
#define NAME_PATH "/eizm/mem/ong1/name"
#define REDUNDANT_NAME_PATH "/eizm/mem/ottmirror/name"
#define TEMPFILE_PATH "/eizm/tmp/ong1/bot3temp"
#define TEMPFILE2_PATH "/eizm/tmp/ong1/bot3temp2"
#define POSTDATA_PATH "/eizm/tmp/ong1/bot3post"
#define COOKIE_PATH "/eizm/tmp/ong1/bot3cook"
#define ONG_STATE_PATH "/eizm/data/aftertime/bftf/ongstate"
#define FORCEMEME_PATH "/eizm/mem/ong1/forcememe"
#define MEME_PATH "/eizm/mem/ong1/mome.txt"

#define WGET_PATH "/usr/bin/wget"
#define MAWK_PATH "/usr/bin/mawk"
#define RM_PATH "/bin/rm"
#define PERL_PATH "/usr/bin/perl"
#define AWK_LATEST "/eizm/pro/ong1/findlatest.awk"
#define AWK_RELATED "/eizm/pro/ong1/related.awk"
#define AWK_POST "/eizm/pro/ong1/post.awk"
#define AWK_MPOST "/eizm/pro/ong1/mpost.awk"
#define AWK_MOME "/eizm/pro/ong1/mome.awk"
#define PERL_MOME "/eizm/pro/ong1/mome.pl"
#define AWK_LASTPOST "/eizm/pro/ong1/lastpost.awk"

int main(int argc,char *argv[]);
int longline(const char *in,const char *out, unsigned long limit);
void end(int m);

unsigned char logopen=0;
FILE *file;
FILE *stopfile;
pid_t sub;
int r;

int main(int argc,char *argv[])
{
	time_t t;
	
	int yarg;
	unsigned char text=0;
	unsigned char otherbot=0;
	unsigned char force=0;
	unsigned char newlog=0;
	unsigned char nopost=0;
	unsigned char mpost=0;
	unsigned long start=0x06666666; // this is a coincidence. It's the largest number that won't overflow the 32bit integer.
	unsigned long last=0xffffffff;
	unsigned long page;
	unsigned long posts;
	
	unsigned short frame=65523;
	unsigned short lastframe=0;
	unsigned char ongstate=0;
	unsigned char pope=0;
	unsigned char selectong=0;
	unsigned char edit;
	
	char name [256];
	char mome[1024];
	char pass [256];
	char bot  [256];
	char argp1[256];
	char argp2[256];
	char argp3[256];
	char argp4[256];
	char argp5[256];
	char URL  [256];
	
	time(&t);
	srand(getpid());
	opterr=0;
	
	while((yarg = getopt(argc,argv,"tnfpmro:")) != -1)
	{
		switch(yarg)
		{
		case 't':
			text=1;
			break;
		case 'n':
			newlog=1;
			break;
		case 'f':
			force=1;
			break;
		case 'p':
			nopost=1;
			break;
		case 'm':
			mpost=1;
			break;
		case 'r':
			otherbot=1;
			break;
		case 'o':
			selectong=1;
			sscanf(optarg,"%hu",&frame);
			ongstate=2;
			break;
		default:
			
			break;
		}
	}
	
	if(!text)
	{
		if(freopen(LOG_PATH,newlog?"wt":"at",stdout)!=NULL)
			logopen=1;
	}
	setvbuf(stdout, NULL, _IONBF, 0);
	dup2(fileno(stdout),fileno(stderr));
	printf("\n%s",ctime(&t));
	
	stopfile=fopen(STOP_PATH,"wt");
	if (stopfile==NULL)
		end(STOP_MSTD);
	if (lockf(fileno(stopfile),F_TLOCK,0)!=0)
		end(STOP_MSTD);	
	fprintf(stopfile,"%hu.FULL STOPFULL STOP\n",0);
		
	if((file=fopen(LAST_PATH,"rt"))!=NULL)
	{
		fscanf(file,"%lu %hu",&start,&lastframe);
		fclose(file);
	}
	printf("start with page %lu, last frame was %hu",start,lastframe);
	
	if(!selectong)
	{
		file=fopen(ONG_STATE_PATH,"rt");
		if(file!=NULL)
		{
			fscanf(file,"%hhu %hu",&ongstate,&frame);
			fclose(file);
		}
		else
			end(LAST_MSTD);
	}
	
	printf(", current frame is %hu",frame);
	
	file=fopen(otherbot?REDUNDANT_NAME_PATH:NAME_PATH,"rt");
	if(file==NULL)
		end(NAME_MSTD);
	if(fgets(argp1,255,file)!=NULL)
		sscanf(argp1,"%s",name);
	else
		end(NAME_MSTD);
	if(fgets(argp1,255,file)!=NULL)
		sscanf(argp1,"%s",pass);
	else
		end(NAME_MSTD);
	if(fgets(argp1,255,file)!=NULL)
		sscanf(argp1,"%s",bot);
	else
		sprintf(bot,"%s","bothasar_p");
	fclose(file);
	printf("\n%s",bot);
	
	if(force==0)
	{
		if(ongstate!=2)
			end(INDELIVERABLE);
		if(frame<=lastframe)
			end(ONGD_MSTD);		
		for(page=start;page<=last;++page)
		{
			if(page!=start)
				sleep(3);
			printf("\n%lu",page);
			sprintf(URL,"%s%lu",OTT_ADDRESS,(page-1)*40);
		
			// download the newpage
			printf(" download");
			sub=fork();
			if(!sub)
			{
				r=execl(WGET_PATH,WGET_PATH, "-q","-t", "3","--connect-timeout=60","-U", USERAGENT, "-O", TEMPFILE_PATH,URL, (char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
			{ 
				sprintf(URL,"%s%lu",TTO_ADDRESS,(page-1)*40);
				printf(" redundant %s",URL);
				sub=fork();
				if(!sub)
				{
					r=execl(WGET_PATH,WGET_PATH, "-q","-t", "3","-U", USERAGENT, "-O", TEMPFILE_PATH,URL, (char *)0);
					exit(r);
				}
				waitpid(sub,&r,0);
				if(r)
				{
					printf(" %d ",r);
					end(DOWN_MSTD);
				}
			}
			printf(" OK, prepare");
			r=longline(TEMPFILE_PATH,TEMPFILE2_PATH,LINE_LIMIT);
			if(r)
				end(PREP_MSTD);
			printf(".");
			
			// what's the last page and how many posts are there?
			sprintf(argp1,"outfile=%s",TEMPFILE_PATH);
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_LATEST,"-v",argp1,TEMPFILE2_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
				end(LAST_MSTD);
			printf(".");
			
			if((file=fopen(TEMPFILE_PATH,"rt"))!=NULL)
			{
				fscanf(file,"%lu %lu",&last,&posts);
				if(posts%40==0)
					pope=1;
				else
					pope=0;
				fclose(file);
			}
			else
				end(LAST_MSTD);
			if(page>last)
				page=last;
			printf(".OK match");
			
			sprintf(argp1,"frame=%hu",frame);
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_RELATED,"-v",argp1,TEMPFILE2_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
			{
				if((file=fopen(LAST_PATH,"wt"))!=NULL)
				{
					fprintf(file,"%lu %hu",last,frame);
					fclose(file);
				}
				else
					end(STOP_MSTD);
				end(ONGD_MSTD);
			}
			else
			{
				if((file=fopen(LAST_PATH,"wt"))!=NULL)
				{
					fprintf(file,"%lu %hu",page,frame);
					fclose(file);
				}
				else
					end(STOP_MSTD);
			}
			printf(" nothing found.");
		}
		
		for(page=1,last=0xffffffff;page<=last;++page)
		{
			if(page!=start)
				sleep(3);
			printf("\nm%lu",page);
			sprintf(URL,"%s%lu",MSTD_ADDRESS,(page-1)*40);
		
			// download the newpage
			printf(" download");
			sub=fork();
			if(!sub)
			{
				r=execl(WGET_PATH,WGET_PATH, "-q","-t", "3","--connect-timeout=60","-U", USERAGENT, "-O", TEMPFILE_PATH,URL, (char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
			{
				printf(" not here.");
				break;
			}
			printf(" OK, prepare");
			r=longline(TEMPFILE_PATH,TEMPFILE2_PATH,LINE_LIMIT);
			if(r)
				break;
			printf(".");
			
			// what's the last page and how many posts are there?
			sprintf(argp1,"outfile=%s",TEMPFILE_PATH);
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_LATEST,"-v",argp1,TEMPFILE2_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
				break;
			printf(".");
			
			if((file=fopen(TEMPFILE_PATH,"rt"))!=NULL)
			{
				fscanf(file,"%lu %lu",&last,&posts);
				if(posts%40==0)
					pope=1;
				else
					pope=0;
				fclose(file);
			}
			else
				break;
			if(page>last)
				page=last;
			printf(".OK match");
			
			sprintf(argp1,"frame=%hu",frame);
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_RELATED,"-v",argp1,TEMPFILE2_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
				end(ONGD_MSTD);
			printf(" nothing found.");
		}
	}	
	
	printf("\nGet meme.");
	
	file=fopen(FORCEMEME_PATH,"rt");
	if(file==NULL)
	{
		sub=fork();
		if(!sub)
		{
			freopen(TEMPFILE_PATH,"wt",stdout);
			r=execl(PERL_PATH,PERL_PATH,PERL_MOME,"-lc","150","-n","67","-d",MEME_PATH,"-t",pope?"[decrall]":"",(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		if(r==0)
		{
			sprintf(argp1,"outfile=%s",TEMPFILE2_PATH);
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_MOME,"-v",argp1,TEMPFILE_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r==0)
			{
				file=fopen(TEMPFILE2_PATH,"rt");
				if(file!=NULL)
				{
					fscanf(file,"%s",mome);
					fclose(file);
				}
				else
					sprintf(mome,"%s","Ong.");
			}
			else
				sprintf(mome,"%s","Ong.");
		}
		else
			sprintf(mome,"%s","Ong.");
	}
	else
	{
		fscanf(file,"%s",mome);
		fclose(file);
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",FORCEMEME_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		if(r)
			sprintf(mome,"%s","Ong.");
	}

	printf(" %s OK.\ncreate post file",mome);
	
	file=fopen(POSTDATA_PATH,"wt");
	if(file != NULL)
	{
		fprintf(file,"username=%s\n",name);
		fprintf(file,"password=%s\n",pass);
		fprintf(file,"subject=bftf-%04hu\n",frame);
		fprintf(file,"addbbcode20=%s\n","100");
		fprintf(file,"message=%s%s%hu%s%04hu%s%s%s%s%s",
			mome,
			"%0A%5Burl%3Dhttp%3A%2F%2F1190.bicyclesonthemoon.info%2Fongoing.htm%3Ff%3D",
			frame,
			((rand()&0x7)?"%5D%5Bimg%5Dhttp%3A%2F%2F1190.bicyclesonthemoon.info%2Fongoing%2Fbftf-":"%5D%5Bimg%5Dhttp%3A%2F%2F1190.bicyclesonthemoon.info%2Fongoing%2Fnpbftf-"),
			frame,
			(frame>=564)?
				".png%5B%2Fimg%5D%5B%2Furl%5D%5Bspoiler%5DI%27d%20highly%20recommend%20following%20the%20first%20commandment%20...again.%5B%2Fspoiler%5D":
				".png%5B%2Fimg%5D%5B%2Furl%5D%5Bspoiler%5D...%5B%2Fspoiler%5D",
			pope?"":"",
			"%5Bsize%3D85%5D%5Bright%5D--%20posted%20by%20",
			bot,
			"%5B%2Fright%5D%5B%2Fsize%5D"
		);
		fclose(file);
	}
	else
		end(POST_MSTD);
	printf(" OK\nsend post");
	
	do
	{//test latest post
		edit=0;
		sub=fork();
		if(sub==0)
		{
			r=execl(WGET_PATH,WGET_PATH, "-q","-t", "3","-U", USERAGENT, "-O", TEMPFILE_PATH,LASTPOST_ADDRESS, (char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		if(r)
			break;
		
		sub=fork();
		if(sub==0)
		{
			if(freopen(TEMPFILE2_PATH,"wt",stdout)==NULL)
				exit(255);
			r=execl(MAWK_PATH,MAWK_PATH, "-f",AWK_LASTPOST, TEMPFILE_PATH, (char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		if(r)
			break;
		
		file=fopen(TEMPFILE2_PATH,"rt");
		if(file==NULL)
			break;
			
		if(fgets(argp1,255,file)!=NULL)
		{
			sscanf(argp1,"%s",argp2);
			sprintf(argp4,"edit=%s",argp2);
		}
		else
		{
			fclose(file);
			break;
		}
		if(fgets(argp1,255,file)!=NULL)
		{
			sscanf(argp1,"%s",argp2);
			sprintf(argp5,"user=%s",argp2);
		}
		else
		{
			fclose(file);
			break;
		}
		fclose(file);
		edit=1;
	}while(0);
	
	sprintf(argp1 ,"postfile=%s",TEMPFILE2_PATH);
	sprintf(argp2,"cookiefile=%s",COOKIE_PATH);
	sprintf(argp3,"tempfile=%s",TEMPFILE_PATH);
	//sprintf(argp4,"name=%s",name);
	//sprintf(argp5,"pass=%s",pass);
	
	if(!nopost)
	{
		if(mpost)
			r=mpost;
		else
		{
			sub=fork();
			if(!sub)
			{
				if(edit)
					r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_POST,"-v",argp1,"-v",argp2,"-v",argp3,"-v",argp4,"-v",argp5,POSTDATA_PATH,(char *)0);
				else
					r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_POST,"-v",argp1,"-v",argp2,"-v",argp3,POSTDATA_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
		}
		if(r)
		{
			sub=fork();
			if(!sub)
			{
				r=execl(MAWK_PATH,MAWK_PATH,"-f",AWK_MPOST,"-v",argp1,"-v",argp2,"-v",argp3,POSTDATA_PATH,(char *)0);
				exit(r);
			}
			waitpid(sub,&r,0);
			if(r)
				end(r);
		}
	}
	printf(" OK");
	end(OK);
	return 0;
}

int longline(const char *in,const char *out, unsigned long limit)
{
	FILE *infile;
	FILE *outfile;
	unsigned char tag;
	unsigned long counter=0;
	char ch;
	
	outfile=fopen(out,"wt");
	if(outfile==NULL)
		return 1;
	infile=fopen(in,"rt");
	if(infile==NULL)
	{
		fclose(outfile);
		return 1;
	}
	
	while(!feof(infile))
	{
		ch=(char)(fgetc(infile));
		fputc(ch,outfile);
		++counter;
		switch(ch)
		{
		case '<':
			tag=1;
			break;
		case '>':
			tag=0;
			if(counter>=limit)
			{
				fputc('\n',outfile);
				counter=0;
			}
			break;
		case ' ':
			if((!tag) && (counter>= limit))
			if(counter>=limit)
			{
				fputc('\n',outfile);
				counter=0;
			}
			break;
		case '\r':
		case '\n':
			counter=0;
			break;
		default:
			break;
		}
		
	}
	return 0;
}


void end(int m)
{
	switch(m)
	{
	case INDELIVERABLE:
		printf("\nCan't ONG this yet.\n");
		break;
	case NAME_MSTD:
		printf("\nName and password fail.\n");
		break;
	case DOWN_MSTD:
		printf("\nDownload fail.\n");
		break;
	case PREP_MSTD:
		printf("\nNewpage prepare fail.\n");
		break;
	case LAST_MSTD:
		printf("\nRemember last time fail.\n");
		break;
	case ONGD_MSTD:
		printf("\nAlready ONGed\n");
		break;
	case POST_MSTD:
		printf("\nPost fail.\n");
		break;
	case STOP_MSTD:
		printf("\nFULL STOPFULL STOP\n");
		break;
	
	
	case OK:
		printf("\nONGED\n");
		break;
	default:
		printf("\nCh*rping m*stard!\n");
	}
	if(logopen)
		fclose(stdout);
	if(m != STOP_MSTD)
	{
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",COOKIE_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",POSTDATA_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",TEMPFILE_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",TEMPFILE2_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		
		lockf(fileno(stopfile),F_ULOCK,0);
		fclose(stopfile);
		
		sub=fork();
		if(!sub)
		{
			r=execl(RM_PATH,RM_PATH,"-f",STOP_PATH,(char *)0);
			exit(r);
		}
		waitpid(sub,&r,0);
		
	}
	exit(m);
}
