#!/usr/bin/perl
#
# mome.pl
# automome
# 09.09.2014
#
# mome.pl is actually mrob27's automome.
# (http://mrob.com/time/automome)
# which he released on the following license:
# #                               LICENSE
# #
# # This software, and the sample memes file that goes with it (which is at
# # http://mrob.com/time/automome/automeme-pl.txt ) are made available under
# # a Creative Commons Attribution-NonCommercial 4.0 International License.
# # License details are at http://mrob.com/cc-license.html

use strict;

# These are required for UTF-8 parsing, which is exemplified by:
#
#   class foreign-words:
#     π
#   regex initial-letter: /^(.).*$/$1/
#   [foreign-words] starts with '[$0/initial-letter/]
#
# however on older OS's, like MacOS 10.4, using these 'use' statements
# sometimes causes false errors from Perl.
use utf8;
use open ':std', ':encoding(UTF-8)';

my $bapropos = q@
automeme(1r)   -- generate random bits of text using templates that match
                  popular internet 'memes'
@;

my $g_memefile = "ott-snowclones.txt";

my $help = qq`
NAME

  automeme     -- generate random bits of text using templates that match
                  popular internet 'memes'

DESCRIPTION

Given no argument, this script just prints out a string of text that
is generated using a randomly-selected template and filling in, after
the style of "Mad Libs".

The meme definitions are in a file called "$g_memefile" in the
current directory or in ~/data/memes ; this can be changed with the -o
and -d options.

OPTIONS

  -d foo/bar.txt  - Read the meme definitions from "foo/bar.txt". You can
                    give any pathname; if nothing is found it prints an error
                    and exits. You can also give just a filename, in which
                    case it will look first in the current directory, then
                    in ~/data/memes
  -lc 60          - Limit number of characters of any output lines to 60
  -n 27           - Make 27 lines of random memey goodness
  -o              - Use the original AUTOMEME snowclone database,
                    "automeme.txt", either in the current directory or in
                    ~/data/memes
  -t '[noun]'     - Instead of selecting templates from those provided
                    in the file, expand the given string as if it were a
                    template.
  -v              - Verbose: Print some debugging messages.

FILE FORMAT

A meme data file can include blank lines or comments starting with
'#', which will be ignored. A line starting with 'class', such as
'class adjective:', starts a list of words/phrases of a particular
category. A line starting with 'templates:' begins the meme templates
(snowclone definitions). The format of meme templates is demonstrated
by the examples in &examples(), which are executed if you pass the
arguments "-n 0".

DEBUGGING

To detect errors in the memes template file, use the command

  automeme -n 10000 -v | grep Error

This tells it to generate a lot of memes and print error messages (-v),
the #grep# will filter out just the error messages.

SEE ALSO

  fortune       - The classic adage generator
  magic-8-ball  - Prognosticator

`;

my $unused_block_header = q`

REVISION HISTORY

 20140406 First version (cloned from #skel#). Get engine working.
Convert @Link's OTT meme database into this grammar.
 20140407 Fix a bug in rep2 index calculation; allow random selection between
two words with [FOO@0.27, BAR]
 20140408 Add -o option.
 20140409 Allow multiple regexps in history substitution; add
'[foo]ms/bar/baz/' generator for defining classes based on other classes.
 20140411 Fix a regexp bug; allow two fixed-probability literals
 20140415 Add cut operator '[]'
 20140427 UTF8 compliance.
 20140505 Add -lc option.
 20140723 Add multiclass syntax.
 20140725 In &rep3, check for the 'LITERAL@0.12, ' parts first, and
allow three such parts. Add more syntax examples to &examples()
  Implement isomorph syntax. $1/foo/bar regexp syntax now allows
'g' modifier.
 20140730 &rep3 now accepts arbitrarily many random literal parts.
Add an explanatory block comment on multiclass.
 20140731 -v option shows an error message if an empty class
is used in a replaceable context.
 20140803 Add some more error detection.
 20140804 Template-expansion code (functions memeify, rep1, rep2 and rep3)
can now call itself recursively.
 20140804 Add named regex syntax (but only for the ms// selector when
building derived classes)
 20140805 Support named regexes in replaceables; add more error
detection. &rep3 now tries to make sure each string acquired from a
call to rep2 is different from any previous ones (within this
particular call to memeify).
 20140806 Add named templates and recursion. Add more error-checking.
 20140807 More error detection
 20140813 More restrictive test for 'class has no members' error
 20140830 Do not need the UTF-8 "use" statements.
 20140907 You can now give fixed probabilities to templates and/or
classes in a rep3 replaceable.
 20140908 Add -t option.

BUGS and TO-DO

A template can be given a name via label, allowing it to be invoked
the same way a class is named (as an option within a replaceable).
Labels must contain no spaces or other odd punctuation; if you want a
literal colon you can use %3a. If you want the template *only* for use
within another template, end its definition with a cut [].
  Invoke a set of labeled templates with "[foo, bar, baz]" which will
cause one to be picked and a recursive call to rep3
  Use a labeled template as a macro/substring of a longer template
with just one label inside brackets e.g. "foo bar [baz]"
  We'll want to be able to do "foo: [bar, baz][]" and invoke [foo] two
or more times in e.g. the Doge meme. This means at least 2 or 3
recursive levels should be allowed.
  Provide a way to indicate that a replaceable should *not* duplicate
an earlier one (or perhaps this should always be true by default for
certain classes of replaceables).
  Use this construct to improve the Doge meme and deal with xkcd #1318

MULTICLASS and ISOMORPHS

A *multiclass* is a set of classes whose members have a one-to-one
correspondence with each other. Here is an example of a multiclass
defining singular and plural nouns:

   multiclass noun[-singular, -plural]:
   CAT     | CATS
   FISH    | FISH
   GOOSE   | GEESE

This defines classes "noun-singular" and "noun-plural". Each noun has
singular and plural forms, called "declensions". It is also useful to
define a multiclass for verbs, in which case we have "conjugations":

   multiclass verb-transitive[-present, -past]:
   CARRY    | CARRIED
   SEE      | SAW
   THROW    | THREW

The general term for declension and conjugation is "inflection", so
each item in a line is a different inflection of the words next to it:
"CATS" is the plural inflection of "CAT" and "SAW" is the past
inflection of "SEE".
    However, a multiclass is useful for things that are not inflections.
Consider the following examples:

   multiclass animal[-singular, -plural, -collective]:
   CAT     | CATS   | CLOWDER
   COW     | HERD   | CATTLE
   FISH    | FISH   | SCHOOL
   GOOSE   | GEESE  | FLOCK

   multiclass vehicle[-singular, -operator]:
   BICYCLE | RIDER
   CAR     | DRIVER
   SHIP    | CAPTAIN
   TRAIN   | ENGINEER

Here we use the multiclass to define *isomorphisms*, such as "FISH is
to SCHOOL as CAT is to _______" (what is the collecive noun for cats?)
and "CAR is to DRIVER as TRAIN is to _______" (what do you call the
driver of a train?). The different words in a row cannot be called
"inflections" so intead I call them "isomorphs".
    To use a multiclass, first invoke a class member using its normal
class name, then get an isomorph using the history variable, a slash,
and the ending of the class name for the isomorph. For example,
the template:

    I SAW A [animal-collective] OF [$0/-plural].

which might generate the sentence "I SAW A SCHOOL OF FISH."

`;

my $hd = $ENV{"HOME"};

my $p_classname = "[-_a-z0-9]+";
my $p_mul_names = "[-, _a-z0-9]+";

my $g_iterations = 1;
my $g_limit_characters = 9999;
my $g_verbose = 0;

my %error_seen;
sub reg_error
{
  my($msg) = @_;
  if ($error_seen{$msg}) {
  } else {
    print "*** Error: $msg\n" if ($g_verbose);
    $error_seen{$msg} = 1;
  }
}

# Check a regexp pattern to determine if it is a valid Perl regexp.
sub validate_regexp
{
  my($pat, $nonfatal) = @_;
  my($evstr);

  if (" $pat)" =~ m/[^\\][\$\@\%][^)]/) {
    # Allow no variables
    return 0;
  }

  if (($pat =~ m/\|/) && ($pat =~ m|/|)) {
    # Cannot handle, fail
    &reg_error("Pattern cannot use both '|' and '/'");
    return 0;
  } elsif ($pat =~ m|/|) {
    $evstr = "'foo' =~ m|$pat|";
  } else {
    $evstr = "'foo' =~ m/$pat/";
  }
  eval $evstr;
  if ($@) {
    # Failage
    # die "val fail /$pat/\n";
    return 0;
  }
  # It must have succeeded
  return 1;
}

sub eval_repl
{
  my($l, $from, $to, $mod) = @_;
  my($rv, $evstr);

  $rv = $l;
  $evstr = '$rv =~ ';
  if (($from =~ m/\|/) && ($from =~ m|/|)) {
    # Cannot handle, just return original string
    return $l;
  } elsif ($from =~ m|/|) {
    $evstr .= "s|$from|$to|$mod;";
  } else {
    $evstr .= "s/$from/$to/$mod;";
  }
  eval $evstr;
  if ($@) {
    # Failage
    # die "repl fail /$from/$to/\n";
    return $l;
  }
  # It must have succeeded
  return $rv;
}

sub eval_match
{
  my($l, $pat, $mod) = @_;
  my($rv, $evstr);

  $rv = 0;
  $evstr = '$rv = ($l =~';
  if ($pat =~ m|/|) {
    $evstr .= "m|$pat|$mod)";
  } else {
    $evstr .= "m/$pat/$mod)";
  }
  eval $evstr;
  if ($@) {
    # Failage
    # die "match fail '$pat'\n";
    return 0;
  }
  # It must have succeeded
  return $rv;
}

my %class_pop;
my %class_member;

my %named_regex;
my %named_template;

# Add line of text to current class
sub add1
{
  my($class, $l) = @_;
  my($key);

  $class_pop{$class} += 0; # Make sure this is not the null string
                           # when defining the first element
  $key = "$class $class_pop{$class}";
  $class_member{$key} = $l;
  # print "class_member{$key} = $class_member{$key}\n";
  $class_pop{$class}++;
}

my $g_mc_width;
my @mc_names;
my @mc_vars;
my %is_multiclass;
my %mc_isomorphs;
my %multiclass_basenames;

# Define the names of each of the classes in a multiclass.
sub mc_setup
{
  my($base, $vars) = @_;
  my($name, $b2, $var);

  $g_mc_width = 0;
  foreach $var (split(/,/, $vars)) {
    $var =~ s/ //g;
    # print "mcs $g_mc_width $var\n";
    $name = $base . $var;
    $mc_names[$g_mc_width] = $name;
    $is_multiclass{$name} = 1;
    $mc_vars[$g_mc_width] = $var;
    $g_mc_width++;
  }

  # Create the isomorph mappings
  for $b2 (@mc_names) {
    for $var (@mc_vars) {
      $mc_isomorphs{"$b2/$var"} = $base . $var;
      # print "mc_isomorphs{'$b2/$var'} = '$base$var'\n";
    }
  }

  $multiclass_basenames{$base} = 1;
}

# Add a list of items to all the classes in the current multiclass. The input
# is a line like "FOO | BAR |    | BAZ" where the items are separated by
# '|'. Any blank items will cause empty strings to be put into the corresponding
# 
sub mc_add
{
  my($l) = @_;
  my($i, $class, $key, $var);

  $i = 0;
  foreach $var (split(/\|/, $l)) {
    if ($i < $g_mc_width) {
      $var =~ s/^ +//;
      $var =~ s/ +$//;
      # print "mcs $i $var\n";
      $class = $mc_names[$i];
      $class_pop{$class} += 0;
      $key = "$class $class_pop{$class}";
      $class_member{$key} = $var;
      # print "class_member{'$key'} = '$var'\n";
      $class_pop{$class}++;
      $i++;
    }
  }
  # Make sure blanks are put in any columns that were left out
  while ($i < $g_mc_width) {
    $class = $mc_names[$i];
    $key = "$class $class_pop{$class}";
    $class_member{$key} = '';
    $class_pop{$class}++;
    $i++;
  }
}

# Load data for vocabulary and snowclone templates
sub load_data
{
  my($dir, $fn, $fp, $l, $class, $key);
  my($vocab, $mclass, $mdim, $p1, $p2);

  # %%% In future the meme database could be selected by an environment
  # variable, command-line option and/or rc file.
  $fn = "$g_memefile";

  # Look for database in curdir or in ~/data/memes
  if ($fn =~ m|^/|) {
    # User gave absolute path
    $fp = $fn;
  } else {
    # Relative path: look in . or in ~/data/memes
    $dir = ".";
    if (!(-f "$dir/$fn")) { $dir = "$hd/data/memes"; }
    $fp = "$dir/$fn";
  }
  die "No file '$fp'\n" if (!(-f $fp));

  print "Reading meme definitions from $fp\n" if ($g_verbose);
  open(my $IN, $fp);
  while($l = <$IN>) {
    chomp $l;
    $l =~ s/[\t]/ /g;
    $l =~ s/^ +//;
    $l =~ s/ +$//;
    $l =~ s/ +/ /g;
    if ($l eq '') {
      # Blank line -- ignore
    } elsif ($l =~ m/^#/) {
      # Comment delimiter -- ignore
    } elsif ($l =~ m/^class +($p_classname):$/i) {
      $class = $1;
      $vocab = 1; $mclass = '';
    } elsif ($l =~ m/^multiclass +($p_classname) *\[($p_mul_names)\]:$/i) {
      $mclass = $1; $mdim = $2;
      # print "mc_setup($mclass, $mdim)\n";
      &mc_setup($mclass, $mdim);
      $vocab = 1; $class = ''; # Suppress adding to simple class
    } elsif ($l =~ m|^regex +($p_classname): *(/.+/.*/[gi]*) *$|i) {
      $p1 = $1; $p2 = $2;
      $named_regex{$p1} = $p2;
    } elsif ($l =~ m|^regex +($p_classname)|i) {
      &reg_error ("Malformed regex definition '$1'");
    } elsif ($l =~ m|^regex +|i) {
      &reg_error ("Malformed regex declaration");
    } elsif ($l =~ m|^template +($p_classname): +(.*)\[\] *$|i) {
      $p1 = $1; $p2 = $2;
      $named_template{$p1} = $p2;
      # print "named_template{$p1} = $p2\n";
    } elsif ($l =~ m|^template |i) {
      &reg_error("'template' keyword without classname, ':' or ending '[]'");
    } elsif ($l =~ m/^templates:$/) {
      $class = 't:';
      $vocab = 0; $mclass = '';
    } elsif ($class ne '') {
      my($fcl, $from, $to, $i, $key, $mod);
      if ($vocab
         && ($l =~ m|^\[($p_classname)\] *ms/([^/]+)/(.*)$|))
      {
        $fcl = $1; $p1 = $2; $p2 = $3;
        if ($named_regex{$p1} ne '') {
          $p1 = $named_regex{$p1};
        } else {
          $p1 = "/$p1/$p2";
        }
        $p1 =~ tr/{}/[]/;
        if ($p1 =~ m|/([^/]+)/([^/]*)/(i?)|) {
          $from = $1; $to = $2; $mod = $3;
          # print "foo1 |$fcl|$from|$to|\n";
          if ($class_pop{$fcl} <= 0) {
            &reg_error("No elements in [$fcl]");
          } elsif (&validate_regexp($from)) {
            for($i=0; $i<$class_pop{$fcl}; $i++) {
              $key = "$fcl $i";
              $l = $class_member{$key};
              if (&eval_match($l, $from, $mod)) {
                # print "foo1 '$l' s/$from/$to/$mod";
                $l = &eval_repl($l, $from, $to, $mod);
                # print "-> '$l'\n";
                &add1($class, $l);
              } else {
                # print "foo4 '$l' does not match /$from/\n";
              }
            }
          } else {
            &reg_error("Could not validate regex /$from/");
          }
        } else {
          &reg_error("Could not parse regex $p1");
        }
      } elsif ($vocab && ($l =~ m|^\[($p_classname)\] *$|))
      {
        $fcl = $1;
        if (($fcl ne $class) && ($class_pop{$fcl} > 0)) {
          for($i=0; $i<$class_pop{$fcl}; $i++) {
            $key = "$fcl $i";
            $l = $class_member{$key};
            &add1($class, $l);
          }
        } elsif ($class_pop{$fcl} <= 0) {
          # print "foo2 no elements in [$fcl]\n";
        } else {
          # print "foo3 could not validate /$from/\n";
        }
      } else {
        &add1($class, $l);
      }
    } elsif ($mclass ne '') {
      &mc_add($l);
    }
  }
  close $IN;
} # End of load_data

# variables used within memeify, and re-instantiated by memeify when
# it calls itself recursively. See www.perlmonks.org/?node_id=66677
our $g_rep2_index;
our $g_rep2_class;

our @g_subst_history;
our $n_meme_substitutions;
our @g_sh_index;
our @g_sh_class;
our @g_sh_oa;
our %rep2_results;

my $complexity;
my $base_template;

# memeify takes a full meme template and expands it. Templates can invoke labeled sub-templates. The final result is returned as a string.
# memeify does not perform the final step of un-escaping %% literals.
sub memeify
{
  my($meme) = @_;

  local $g_rep2_index;
  local $g_rep2_class;

  local @g_subst_history;
  local $n_meme_substitutions;
  local @g_sh_index;
  local @g_sh_class;
  local @g_sh_oa;
  local %rep2_results;

  # print "memeify '$meme', n_meme_substitutions == $n_meme_substitutions\n";
  # print "foo1 gsh[0] == $g_subst_history[0]\n";

  # Return a single substitution from a given class, chosen at random
  sub rep1
  {
    my($class) = @_;
    my($i, $t);

    $i = int($class_pop{$class} * rand);
    $t = $class_member{"$class $i"};
    return $t;
  }

  # rep2 takes a list of classes OR a list of named templates, and either
  # selects a class member or selects and executes a template.
  #
  # In the first form, it returns a substitution chosen from one or more
  # classes given a list of those classes' names. The distribution
  # will be even by individual class-member, not by class. For
  # example, if you give the classes "fruits, vegetables" and if there
  # are 14 vegetables and 11 fruits, then this function will return a
  # vegetable 14/(14+11) = 56% of the time.
  #
  # In the second form, it chooses a named template given a list of template
  # names, then executes the selected template by making a single recursive
  # call to memeify. In this instance, each named template has an equal
  # probability of being chosen.
  sub rep2
  {
    my($classes) = @_;
    my($cls, $tpop, $ncls, $i, $na, $nr, $recurse);
    my(@bases);
    my(@cnames);

    $classes =~ s/,/ /g;
    $classes = " $classes ";
    $ncls = 0; $tpop = 0; $recurse = 0;
    foreach $cls (split / /, $classes) {
      if ($cls =~ m/^$p_classname$/) {
        if ($class_pop{$cls}) {
          if ($recurse) {
            &reg_error("Cannot mix named templates and literal-classes in a pick-list.");
          } else {
            $cnames[$ncls] = $cls;
            $bases[$ncls] = $tpop;
            $tpop += $class_pop{$cls};
            $ncls++;
          }
        } elsif ($multiclass_basenames{$cls}) {
          &reg_error("'$cls' is a multiclass basename, variant missing.");
        } elsif ($named_template{$cls} ne '') {
          if ($tpop > 0) {
            &reg_error("Cannot mix literal-classes and named templates in a pick-list.");
          } else {
            $recurse = 1;
            $cnames[$ncls] = $cls;
# print "cnames[$ncls] = '$cls'\n";
            $ncls++;
          }
        } elsif (($cls =~ m/[a-z]/) && ($cls eq lc($cls))) {
          &reg_error("Class '$cls' has no members.");
        }
      }
    }

    if ($recurse) {
      # Pick one at random
      $na = int($ncls * rand);
      # Retrieve the template and memeify it!
      $cls = &memeify($named_template{$cnames[$na]});
      return $cls;
    }

    if ($ncls == 0) {
      # We found none, so we'll just return the string
      $classes =~ s/^ *//;
      $classes =~ s/ *$//;
      return $classes;
    }

    $na = int($tpop * rand);

    # Find out what class we selected
    $i=0; while (($na>=$bases[$i+1]) && ($i+1 < $ncls)) { $i++; }

    # Convert the absolute index into an index relative to the selected class
    $nr = $na - $bases[$i];
    $cls = $cnames[$i];

    # If we're drawing from a member of a multiclass, it makes sense to
    # remember the class name and index for use in later isomorph
    # substitution
    if ($is_multiclass{$cls}) {
      $g_rep2_index = $nr;
      $g_rep2_class = $cls;
    } else {
      $g_rep2_index = -1;
      $g_rep2_class = '';
    }

    # print "foo1 $cls: ncls $ncls tpop $tpop na $na nr $nr\n";

    # We're done!
    return $class_member{"$cls $nr"};
  } # End of rep2

  # Call rep2 with a repetition-avoidance loop. Results are considered
  # 'duplicate' if the result (output of rep2) is identical to a
  # previous rep2 result, regardless of whether the argument (input to
  # rep2) is the same or different.
  sub avoid_rep2
  {
    my($arg) = @_;

    # We iteratively call &rep2 until it manages to return a result that
    # is distinct from any other &rep2 calls within this instance of
    # &memeify. To avoid infinite looping, we try at most 27 times.
    my($avoid_count) = 27;
    my($rv);

    while ($avoid_count > 0) {
      $rv = &rep2($arg);
      if (($rep2_results{$rv}+0) == 0) {
        # We found a new one!
        $avoid_count = 0;
      } else {
        $avoid_count--;
      }
    }
    $rep2_results{$rv} = 1;

    return $rv;
  }

  # Return a replacement for one of the types of substitutions defined by the
  # automeme grammar (see FILE FORMAT section in block comment above, and
  # examples in &examples() routine below)
  sub rep3
  {
    my($arg) = @_;
    my ($i, $rv, $from, $to, $rs, $mod, $gh_index, $gh_class, $oa, $t);

    # print "foo2 gsh[0] == $g_subst_history[0]\n";

    $rv = ''; $oa = $arg;
    $gh_index = -1; $gh_class = '';
    if ($arg eq '') {
      # Cut: send cut signal back to caller
      return '[]';
    }

    $i = 0;
    while ($arg =~ s/^([^\$\@]*)\@([.0-9]+), *//) {
      if ((rand() < $2) && ($i == 0)) {
        # Return initial string (a template, class, or literal)
        # with a fixed probability
        $rv = $1;
        if (($class_pop{$rv} > 0) || ($named_template{$rv} ne '')) {
          $rv = &avoid_rep2($rv);
        }
        $i = 1;
      }
    }

    if ($i) {
      # We got one of the randoms
    } elsif ($arg =~ m|^\$([0-9]+) *(/.*/.*) *$|) {
      # numbered 'variable' with (one or more) regexp replacement(s)
      # demarcated by at least two slash characters '/'
      $rv = $1; $rs = $2;
      # Check for named regex
      if ($rs =~ m|^/($p_classname)/$|) {
        # With just two slashes, we expect a named regex.
        $t = $1;
        if ($named_regex{$t} ne '') {
          $rs = $named_regex{$t};
        } elsif ($g_sh_class[$i] ne '') {
          # Perhaps they're trying to do an isomorph
          &reg_error("I got '/$t/', but isomorph syntax has no trailing slash.");
        } else {
          &reg_error("No named regex '$t'");
        }
      } else {
        # This is okay, it's probably a literal regex.
        # &reg_error("'$rs' is not of the named-regex form");
      }
      if ($rv >= $n_meme_substitutions) {
        &reg_error("There is no '\$$rv' in history.");
      }
      $rv = $g_subst_history[$rv];
      $rs .= ', ';
      # print "rep3 |$arg|$rv|$rs|\n";
      # Bite off the grep-forms one at a time. Each is expected to
      # be of the form /foo/bar/x where the match 'foo' must be non-null,
      # and the trailing modifiers 'x' can include 'g' or 'i' but nothing
      # else.
      while ($rs =~ m|^/([^/]+)/([^/]*)/([gi]*), *(.*)$|) {
        $from = $1; $to = $2; $mod = $3; $rs = $4;
        $from =~ tr/{}/[]/;
        # print "rep3 '$rv' /$from/$to/";
        if (&validate_regexp($from)) {
          $rv = &eval_repl($rv, $from, $to, $mod);
        } else {
          &reg_error("Could not validate regex /$from/");
        }
        # print " -> '$rv'\n";
      }
      # See if we are left with anything; this is a syntax error.
      if ($rs ne '') {
        &reg_error("Malformed or partial regex '$rs'");
      }
    } elsif ($arg =~ m|^\$([0-9]+) */([^/]+)$|) {
      # numbered 'variable' with isomorph substitution
      $i = $1; $rs = $2;
      # print "foo $i $g_sh_class[$i]\n";
      if ($g_sh_class[$i] eq '') {
        # We cannot isomorph, just return the original
        $rv = $g_subst_history[$i];
        if ($named_regex{$rs} ne '') {
          &reg_error("In '[\$$i/$rs]', the named regex needs a trailing slash.");
        } else {
          &reg_error("'$g_sh_oa[$i]' is not a multiclass member.");
        }
      } else {
        # Get isomorph class
        $rv = $mc_isomorphs{"$g_sh_class[$i]/$rs"};
        # print "mc_isomorphs{'$g_sh_class[$i]/$rs' == $rv\n";
        if ($rv eq '') {
          # They asked for a nonexistent isomorph
          $rv = $g_subst_history[$i];
          # print "foo-err2\n";
          if ($named_regex{$rs} ne '') {
            &reg_error("Named regex '$rs' needs a trailing slash.");
          } else {
            &reg_error("'$g_sh_class[$i]' has no isomorph '$rs'.");
          }
        } else {
          $rv = $class_member{"$rv $g_sh_index[$i]"};
        }
      }
    } elsif ($arg =~ m|^\$([0-9]+) *$|) {
      # numbered 'variable': return the Nth substitution from those we have
      # generated so far for this meme. We return directly so as not to
      # define another history item, which would be redundant.
      return $g_subst_history[$1];
    } elsif ($arg =~ m|^\$([0-9]+)(.*)$|) {
      &reg_error("Unknown modifier '$2' on history variable");
    } else {
      # default case: $arg contains one or more word classes
      # implementation differs depending on whether equal weight is given per
      # class or per class-member.
      # %%% recursion passes through here; callers checked in rep2

      $rv = &avoid_rep2($arg);

      $gh_index = $g_rep2_index;
      $gh_class = $g_rep2_class;
    }

    $g_subst_history[$n_meme_substitutions] = $rv;
    $g_sh_index[$n_meme_substitutions] = $gh_index;
    $g_sh_class[$n_meme_substitutions] = $gh_class;
    $g_sh_oa[$n_meme_substitutions] = $oa;
    $n_meme_substitutions++;

    return $rv;
  } # End of rep3

  $complexity++;
  if ($complexity > 27) {
    &reg_error("Recursive template too complex: $base_template");
    return "";
  }

  # Do it!
  $meme =~ s/\[([^]]+)\]/&rep3($1)/ge;

  # Detect a cut signal if any
  $meme =~ s/^.*\[\]//;

  return $meme;
} # End of memeify

sub mfy2
{
  my($template) = @_;
  my($rv);

  $complexity = 0;
  $base_template = $template;

  $rv = &memeify($template);

  # De-escape any special literals
  $rv =~ s/\%([0-9a-fA-F]{2})/chr(hex($1))/ge;

  return $rv;
}

sub examples
{
  # Pass "-n 0" as arguments to see these demonstration test cases

  # Return an adjective
  print (&mfy2("[adjective]") . "\n");

  # Return a present-continuous (aka present progressive) or past-tense verb
  print (&mfy2("[verb-tr-cont, verb-tr-past]") . "\n");

  # Return a random noun, but half the time it's HODOR. In Perl
  # @ inside "" is special so we use ''
  print (&mfy2('[HODOR@0.5, noun-singular]') . "\n");

  # Return NORTH 1/4 of the time, EAST 1/3 of the remaining time,
  # SOUTH 1/2 of the remaining time, and WEST the rest of the time.
  print (&mfy2('[NORTH@0.25, EAST@0.3333, SOUTH@0.5, WEST]') . "\n");

  # Example of a full expansion
  print (&mfy2('I CAN [verb-tr-inf] [HOTDOG@0.3, noun-singular] PLEEZ?') . "\n");

  # This one uses the % escape character to include literal brackets [] in the result
  print (&mfy2('%5b[CITATION@0.3, noun-singular, noun-mass] [NEEDED@0.3, verb-tr-past]%5d') . "\n");

  # A few examples that use history substitition
  print (&mfy2('I PUT [noun-mass] IN YOUR [$0]...') . "\n");

  # This snowclone performs a sort of Pig Latin transformation. The prototype
  # is "Bless you, ants. Blants." but we want it to work on plural nouns
  # that start with a consonant. We use a regexp to remove any string of zero
  # or more non-vowels from the beginning of $0. Note the use of braces {}
  # to represent a character-class within the regexp.
  print (&mfy2('BLESS YOU, [noun-plural, noun-mass, character]. BL[$0/^{^AEIUO}+//].') . "\n");

  # Here we use a regexp to remove the article (or any initial word) from a
  # noun-singular-article
  print (&mfy2('I HAZ [noun-singular-article]. U CAN HAS MY [$0/^{^ }+ //].') . "\n");

  # For the "Dial M for murder" template we need to choose the full word
  # 'murder' before being able to extract its first letter. To achieve this
  # we use the cut operation '[]', which causes the parser to evaluate
  # the preceding portion and then discard it.
  print (&mfy2('[noun-singular, noun-plural, noun-mass, character] []DIAL [$0/^(.).*/$1/] FOR [$0]') . "\n");

  # Another example, where we select a [place] and then remove any initial 'THE'
  print (&mfy2('[place][]PH\'NGLUI MGLW\'NAFH [character] [$0/^THE //] WGAH\'NAGL FHTAGN') . "\n");

  # A sometimes flawed attempt at verb conjugation. The prototype is
  # "Hating haters gonna hate", and we attempt to transform the continuous
  # verb form ("hating") into the plural actor form ("haters") by matching the
  # ING and changing it to ERS. We than attempt to form the infinitive ("hate")
  # by simply removing ING, but that doesn't always work. Here we show a few
  # exceptions
  print (&mfy2('[verb-tr-cont, verb-intr-cont] [$0/YING/IERS/, /ING/ERS/] GONNA [$0/ING//, /LLAT$/LLATE/, /DOODL /DOODLE /, /HOTDOGG/HOTDOG/]') . "\n");

  # Another attempt at verb transformation
  print (&mfy2('[verb-intr-cont][verb-intr-cont][]THEY SEE ME [$0/ING/IN\'/], THEY [HATIN\'@0.67, $1/ING/IN\'/]') . "\n");

  print (&mfy2('I WEAR [noun-mineral-art] NOW. [$0/-plural] ARE COOL.') . "\n");

  print (&mfy2('[noun-animal-plural, noun-mineral-plural][$0/-art][$1/^{^ }+ //][][$0] FOR THE [$2] GOD!') . "\n");

  # Demonstrating use of a meme subroutine ('[verbmeme]'), buth the main meme and the subroutine use $0, and the main meme extracts parts of the subroutine.
  print (&mfy2('[character][character][][$1] TURNED TO LOOK AT [$0]. I ASKED [$0], AND XE ANSWERED, "[verbmeme]". "REALLY, [$0]?" I REPLIED, "[$2/^.* ({^ }+)$/$1/]? WHY NOT [verb-tr-inf] THEM?" [$1] LOOKED BACK AT ME AND REPLIED, "BECAUSE [$4/-cont] [$2/^.* ({^ }+):.*$/$1/] IS [****DISH@0.5, adjective-negative]."') . "\n");

  # Demonstrating the avoidance algorithm.
  print (&mfy2("[avoid] [avoid] [avoid] [avoid]") . "\n");
}

$| = 1;

# Set defaults and parse arguments
my ($arg, $i, $g_literal_template);
while($arg = shift) {
  if ($arg =~ m/^-[-]?h(elp)?$/) {
    print $help;
    exit(0);
  } elsif ($arg eq '-d') {
    # Use user-specified memebase
    $g_memefile = shift;
  } elsif ($arg eq '-lc') {
    # Limit characters: max length for memes
    $g_limit_characters = shift;
  } elsif ($arg eq '-n') {
    # Produce N memes
    $g_iterations = shift;
  } elsif ($arg eq '-o') {
    # Use original memebase
    $g_memefile = "automeme.txt";
  } elsif ($arg eq '-t') {
    $g_literal_template = shift;
  } elsif ($arg eq '-v') {
    $g_verbose = 1;
  } else {
    print STDERR "Unrecognized argument '$arg'\n";
    exit(-1);
  }
}

# Read in the memes database
&load_data();

if ($g_iterations <= 0) {
  &examples();
}

# Produce N memes
my ($m, $gg);
for($i=0; $i<$g_iterations; $i++) {
  # Get memes until we have one that's short enough
  $gg = 1;
  while($gg) {
    if ($g_literal_template ne '') {
      $m = &mfy2($g_literal_template);
    } else {
      $m = &mfy2(&rep1('t:'));
    }
    $gg++;
    if ($gg > 10000) {
      print STDERR "I have tried 10000 times without finding a short enough result.\n";
      exit(-1);
    }
    if (length($m) < $g_limit_characters) {
      $gg = 0;
    }
  }
  print ("$m\n");
}
